#ifndef _LINUX_AIO_H
#define _LINUX_AIO_H
/*
 * A kernel implementation of POSIX Asynchronous I/O facility (KAIO).
 *
 * Copyright 1999, Silicon Graphics, Inc.
 *
 * Written October 1999 by Rajagopal Ananthanarayanan (ananth)
 *  			at Silicon Graphics, Inc.
 */

/*
 * User code gets access to AIO by simply including this header file.
 * CONFIG_AIO has to be explicitly turned on for the kernel.
 */

#ifndef __KERNEL__
#define CONFIG_AIO
#endif

#if defined(CONFIG_AIO)

#ifdef __KERNEL__
#include <asm/siginfo.h>
#else
#include <sys/types.h> 		/* off_t, etc. 				*/
#include <signal.h>		/* sigevent_t, etc. 			*/
#endif

/*
 * Default & maximum number of threads to be used for processing
 * parts of aio request. Number of threads can be changed using the
 * environment variable below.
 */

#define AIO_THREAD_ENVIRON	"AIO_NTHREADS"
#define AIO_DEFAULT_THREADS 	 4
#define AIO_MAX_THREADS		64

#define AIO_TIMES	8

/*
 * User code can turn on 64-bit offsets in aiocb by:
 * 	#define _FILE_OFFSET_BITS 64
 *	#include <features.h>
 *	#include <linux/aio.h>
 * which will turn all aio_*() interfaces to operate with 64-bit offsets.
 * 
 * Further, #define _LARGEFILE64_SOURCE will enable LFS style aio_*64()
 * interfaces. Important note: <features.h> should be included as
 * the first file in a code module, for the above to work correctly.
 * See <features.h> for further details.
 *
 * An alternate way to turn on 64-bit offsets in aiocb is to:
 *	#define AIO_FILE_OFFSET64
 *	#include <linux/aio.h>
 * Similarly LFS interfaces can be enabled by:
 *	#define AIO_LARGEFILE64
 *	#include <linux/aio.h>
 * 
 * The kernel always uses 64-bit offset values. If user code does not
 * enable 64-bit offsets, the library will suitably zero out the padding.
 */
typedef struct aiocb {
	int		aio_fildes;	/* File descriptor to perform aio  */ 
	void		*aio_buf;	/* User's data buffer 		   */
	size_t		aio_nbytes;	/* No. of bytes in transfer 	   */
#if defined(__USE_FILE_OFFSET64) || defined(AIO_FILE_OFFSET64) || defined(__KERNEL__)
	loff_t		aio_offset;
#else
	off_t		aio_offset;	/* File position to begin transfer */
	char		__aio_pad[sizeof(loff_t) - sizeof(off_t)];
#endif
	int		aio_reqprio;	/* Unused - Reserved 		   */
	sigevent_t	aio_sigevent;	/* Notification Options 	   */
	int		aio_lio_opcode; /* lio_listio opcode 		   */
	unsigned long	aio_reserved[7];/* Don't use this 		   */
	unsigned long	aio_times[AIO_TIMES];
} aiocb_t;


#if defined(__USE_LARGEFILE64) || defined(AIO_LARGEFILE64)
typedef struct aiocb64 {
	int		aio_fildes;	/* File descriptor to perform aio  */ 
	void		*aio_buf;	/* User's data buffer 		   */
	size_t		aio_nbytes;	/* No. of bytes in transfer 	   */
	loff_t		aio_offset;	/* File offset - 64 bits	   */
	int		aio_reqprio;	/* Unused - Reserved 		   */
	sigevent_t	aio_sigevent;	/* Notification Options 	   */
	int		aio_lio_opcode; /* lio_listio opcode 		   */
	unsigned long	aio_reserved[7];/* Don't use this 		   */
	unsigned long	aio_times[AIO_TIMES];
} aiocb64_t;
#endif

/*
 * Internal to the aio library
 */
#define AIOCMD_READ		1
#define AIOCMD_SUSPEND		2
#define AIOCMD_CANCEL		3
#define AIOCMD_RETURN		4
#define AIOCMD_PCINVALIDATE	5	/* invalidate page cache	   */
#define AIOCMD_SLAVE		6	/* put a slave in wait for request */
#define AIOCMD_WRITE		7
#define AIOCMD_GENERIC_READ	8 	/* Reserved for internal use	   */
#define AIOCMD_LIST_IO		9

/*
 * For gathering timing statistics.
 */
#define AIOTIME_REQUEST		0
#define AIOTIME_IO		1
#define AIOTIME_COPY		2

#if !defined(__USE_FILE_OFFSET64) && !defined(AIO_FILE_OFFSET64)
/*
 * Procedural interfaces.
 */
int 	aio_read(struct aiocb *aiocbp);
int 	aio_write(struct aiocb *aiocbp);
ssize_t	aio_return(struct aiocb *aiocbp);
int	aio_cancel(int fildes, struct aiocb *aiocbp);
int	aio_fsync(int op, struct aiocb *aiocbp);
int	aio_suspend(const struct aiocb *const list[],
			 int nent, const struct timespec *timeout);
int 	lio_listio(int mode, struct aiocb *const list[],
			int nent, struct sigevent *sig);
#else
# ifdef __REDIRECT
extern int __REDIRECT (aio_read, __P ((struct aiocb *__aiocbp)), aio_read64);
extern int __REDIRECT (aio_write, __P ((struct aiocb *__aiocbp)), aio_write64);
extern int __REDIRECT (lio_listio, __P ((int __mode,
		     struct aiocb *__const __restrict __list[],
		     int __nent, struct sigevent *__restrict __sig)), lio_listio64);
extern __ssize_t __REDIRECT (aio_return, __P ((struct aiocb *__aiocbp)), aio_return64);
extern int __REDIRECT (aio_cancel, __P ((int __fildes, struct aiocb *__aiocbp)),
		       aio_cancel64);
extern int __REDIRECT (aio_suspend,
		       __P ((__const struct aiocb *__const __restrict __list[],
			     int __nent,
			     __const struct timespec *__restrict __timeout)),
		       aio_suspend64);
#else
#define aio_read aio_read64
#define aio_write aio_write64
#define lio_listio lio_listio64
#define aio_return aio_return64
#define aio_cancel aio_cancel64
#define aio_suspend aio_suspend64
#endif
#endif

#define aio_error(a)	(volatile int)(a)->aio_reserved[2]
/*
 * 64-bit interfaces.
 */
#if defined(__USE_LARGEFILE64) || defined(AIO_LARGEFILE64)
int 	aio_read64(struct aiocb64 *aiocbp);
int 	aio_write64(struct aiocb64 *aiocbp);
ssize_t	aio_return64(struct aiocb64 *aiocbp);
int	aio_cancel64(int fildes, struct aiocb64 *aiocbp);
int	aio_fsync64(int op, struct aiocb64 *aiocbp);
#define aio_error64(a)	aio_error(a)
int	aio_suspend64(const struct aiocb64 *const list[],
			 int nent, const struct timespec *timeout);
int 	lio_listio64(int mode, struct aiocb64 *const list[],
			int nent, struct sigevent *sig);
#endif

/*
 * Return values for aio_cancel.
 */
#define AIO_CANCELED		1
#define AIO_NOTCANCELED		2
#define AIO_ALLDONE		3

/*
 * Listio modes
 */
#define LIO_WAIT		1
#define LIO_NOWAIT		2


/*
 * Listio opcodes
 */
#define LIO_READ		1
#define LIO_WRITE		2
#define LIO_NOP			3

#ifdef __KERNEL__

#if defined(CONFIG_RAW)
#include <linux/iobuf.h>
/*
 * Interfaces expected from Raw I/O for AIO.
 */
extern	int kaio_is_raw(struct file *);
extern	int kaio_raw_wait(void *, struct kiobuf *);
extern	void *kaio_raw_rw(int, struct file *, char *, size_t, loff_t *, 
			  int *, struct semaphore *, struct kiobuf *);
#endif

typedef struct liocb {
	spinlock_t 	 lio_lock; 	/* protects nent manipulation	   */
	int	   	 lio_nent; 	/* number of kaio's being awaited  */
	int	   	 lio_done; 	/* number of kaio's completed	   */
	struct wait_queue  *lio_wait;	/* wait here for completion of lio */
	struct sigevent	   lio_sigevent;/* completeion signal info, if any */
	struct task_struct *lio_task;	/* requesting task		   */
	int		 lio_mode;	/* mode used in request 	   */
} liocb_t;

typedef struct kaiocb {

	int		kaio_fildes;	/* file to operate on */
	void		*kaio_buf;	/* user's data buffer */
	size_t		kaio_nbytes;
	loff_t		kaio_offset;
	sigevent_t	kaio_sigevent;	/* Notification Options		*/
	int 		kaio_cmd;	/* read / write 		*/
	int		kaio_error;
	int		kaio_id;	/* unique id for this kaiocb	*/
	struct kaiocb	*kaio_h_next;	/* hash table pointer		*/
	struct kaiocb	**kaio_h_pprev;	/* hash table pointer		*/

	struct page 	**kaio_pages;	/* list of kernel pages   	*/
	int		kaio_npages; 	/* no. of kernel pages 	  	*/
	int		kaio_pchits; 	/* STATS: page cache hits 	*/
	atomic_t	kaio_pending;	/* no. of pages pending I/O	*/
	struct wait_queue  *kaio_wait;	/* wait here if pending I/O	*/
	struct wait_queue  **kaio_suspend_wait;	/* wait here on kaio_suspend */
	struct list_head   kaio_ioq;	/* pointers for queueing aio 	*/
	struct task_struct *kaio_task;	/* requesting task		*/
	void		   *kaio_kq;	/* kaioq  of requesting task	*/
	struct file	   *kaio_filp;	/* held by an fget() 		*/
	aiocb_t		   *kaio_uaiocb; /* user aiocb for this request  */

#if defined(CONFIG_RAW)
	/* RAW I/O fields */
	struct semaphore kaio_sem;	/* for use by raw I/O		*/
	struct kiobuf 	 kaio_iobuf; 	/* for use by raw I/O		*/
	void 		 *kaio_scp;	/* cookie returned by raw I/O   */
#endif

	liocb_t		*kaio_liocb; 	/* List I/O control block pointer */

#ifdef KAIO_STATS
	unsigned long	kaio_times[AIO_TIMES];
#define kaio_set_time(k, x) 	(k)->kaio_times[(x)] = kaio_time()
#else
#define kaio_set_time(k, x)
#endif /* KAIO_STATS */

} kaiocb_t;


typedef struct kaio_key {
	int	kaio_kid;
	void	*kaio_kq;
} kaio_key_t;


#define aio_key		aio_reserved[0]
/* 	aio_key[1]	aio_reserved[1] */

#undef aio_error	/* don't want user definition which is a function */
#define aio_error	aio_reserved[2]
#define aio_return	aio_reserved[3]

extern int kaio_rw(int cmd, struct aiocb *, liocb_t *);
extern int kaio_list_io(int, struct aiocb *[], int, sigevent_t *);
extern int kaio_suspend(aiocb_t *[], int, struct timespec *, int);
extern int kaio_cancel(int, kaio_key_t *);
extern void kaio_init(void);
extern void kaio_slave(void);
extern void kaio_q_create(struct task_struct *);


/*
 * Prefix for the names of slave threads
 */
#define KAIO_SLAVE_NAME		"kaiod-"
#define KAIO_SLAVE_NAMELEN	6

#endif /* __KERNEL__ */

#endif /* CONFIG_AIO */

#endif /* _LINUX_AIO_H */
